##############################################################################
#
# Kultur Tabbed Abstract Page Renderer
#
##############################################################################

# Copyright 2009 Tim Brody <tdb2@ecs.soton.ac.uk>, University of Southampton
#
# This file is released under the GPLv3 license.
#
# Usage (in cfg.d/eprint_render.pl):
#
# $c->{eprint_render} = sub {
# 	...
# 	return kultur_eprint_render( $session, $eprint, $preview );
# };

######################################################################

=item $xhtmlfragment = kultur_eprint_render( $eprint, $session, $preview )

This method renders an XHTML fragment that describes the EPrint and contained documents. This generates a tabbed view.

=cut

######################################################################

sub kultur_eprint_render
{
	my( $session, $eprint, $preview ) = @_;

	my $dataset = $eprint->get_dataset;

	my $page = $session->make_doc_fragment;

	my @docs = $eprint->get_all_documents;

	my $header = $session->make_doc_fragment;
	my $content = $session->make_doc_fragment;
	my $footer = $session->make_doc_fragment;

	# render eprint citation
	&kultur_render_citation( $header, $footer, $eprint, \@docs );

	# render the version tree
	&kultur_render_succeeds( $header, $footer, $eprint, \@docs );

	# render the commentary tree
	&kultur_render_commentary( $header, $footer, $eprint, \@docs );

	# render request-a-copy
	&kultur_render_request( $header, $footer, $eprint, \@docs );

	# render editor edit link
	unless( $preview )
	{
		&kultur_render_edit_link( $header, $footer, $eprint, \@docs );
	}

	my @tabs = (
		# render document tab(s)
		&kultur_render_documents( $session, $dataset, $eprint, \@docs ),
		# render metadata tab(s)
		&kultur_render_metadata( $session, $dataset, $eprint, \@docs ),
	);

#	my %box_options;
#	$box_options{session} = $session;
#	$box_options{id} = 'metadata';
#	$box_options{title} = $session->html_phrase( "page:metadata" );
#	$box_options{content} = $table;
#	$box_options{collapsed} = 1;

#	$page->appendChild(EPrints::Box::render(%box_options)); 

	my $id_prefix = "eprint.".$eprint->get_id;

	# render the tab menu
	$content->appendChild( $session->render_tabs(
		id_prefix => $id_prefix,
		current => $tabs[0]->{id},
		tabs => [map { $_->{id} } @tabs],
		labels => {map { $_->{id} => $_->{title} } @tabs},
		links => {map { $_->{id} => "" } @tabs},
		) );

	# panel that all the tab content sits in
	my $panel = $session->make_element( "div",
		id => "${id_prefix}_panels",
		class => "ep_tab_panel",
	);
	$content->appendChild( $panel );

	# populate each of the panels
	foreach my $tab (@tabs)
	{
		my $view_div = $session->make_element( "div",
			id => "${id_prefix}_panel_".$tab->{id},
		);
		if( $tab ne $tabs[0] )
		{
			$view_div->setAttribute( "style", "display: none" );
		}
		$view_div->appendChild( $tab->{content} );

		$panel->appendChild( $view_div );
	}

	$page->appendChild( $header );
	$page->appendChild( $content );
	$page->appendChild( $footer );

	# page title
	my $title = $eprint->render_description();

	# meta links to add to page <head>
	my $links = &kultur_render_links( $session, $dataset, $eprint, \@docs );
	
	return( $page, $title, $links );
}

# render the eprint citation
sub kultur_render_citation
{
	my( $header, $footer, $eprint ) = @_;

	my $session = $eprint->{session};

	my $div = $session->make_element( "div", class=>"ep_block" );
	$header->appendChild( $div );
	$div->appendChild( $eprint->render_citation() );
}

# add a link to the edit-page for this record. Handy for staff.
sub kultur_render_edit_link
{
	my( $header, $footer, $eprint ) = @_;

	my $session = $eprint->{session};

	my $div = $session->make_element( "div", class => "ep_block", style => "text-align: right" );
	$footer->appendChild( $div );
	$div->appendChild( $session->html_phrase( "page:edit_link",
		link => $session->render_link( $eprint->get_control_url ) ) );
}

# render the version tree
sub kultur_render_succeeds
{
	my( $header, $footer, $eprint ) = @_;

	my $succeeds_field = $eprint->get_dataset->get_field( "succeeds" );
	return unless $eprint->in_thread( $succeeds_field );

	my $session = $eprint->{session};

	# Put in a message describing how this document has other versions
	# in the repository if appropriate
	my $div = $session->make_element( "div", class=>"ep_block" );
	$header->appendChild( $div );
	my $latest = $eprint->last_in_thread( $succeeds_field );
	if( $latest->get_value( "eprintid" ) == $eprint->get_value( "eprintid" ) )
	{
		$div->appendChild( $session->html_phrase( "page:latest_version" ) );
	}
	else
	{
		$div->appendChild( $session->render_message(
			"warning",
			$session->html_phrase( "page:not_latest_version",
				link => $session->render_link( 
						$latest->get_url() ) ) ) );
	}

	# the version thread
	$div = $session->make_element( "div", class=>"ep_block" );
	$footer->appendChild( $div );
	$div->appendChild( $session->html_phrase( "page:available_versions" ) );
	$div->appendChild( $eprint->render_version_thread( $succeeds_field ) );
}

# render the commentary tree
sub kultur_render_commentary
{
	my( $header, $footer, $eprint ) = @_;

	my $commentary_field = $eprint->get_dataset->get_field( "commentary" );
	return unless $eprint->in_thread( $commentary_field );

	my $session = $eprint->{session};

	# the commentary thread
	my $div = $session->make_element( "div", class=>"ep_block" );
	$footer->appendChild( $div );
	$div->appendChild( $session->html_phrase( "page:commentary_threads" ) );
	$div->appendChild( $eprint->render_version_thread( $commentary_field ) );
}

# render the eprint metadata table
sub kultur_render_metadata
{
	my( $session, $dataset, $eprint ) = @_;

	my( $table, $tr, $td, $th );	# this table needs more class cjg
	$table = $session->make_element( "table",
					class=>"ep_block",
					border=>"0",
					cellpadding=>"3" );

	# KULTUR - show additional metadata
	for( qw( creators_name corp_creators ) )
	{
		next unless $eprint->is_set( $_ );
		$table->appendChild( $session->render_row(
			$session->html_phrase( "eprint_fieldname_$_" ),
			$eprint->render_value( $_ ) ) );
	}

	#abstract
	if ( $eprint->is_set( "abstract" ) )
	{
		$table->appendChild( $session->render_row(
					$session->html_phrase( "eprint_fieldname_abstract" ),
					$eprint->render_value( "abstract" ) ) );
	}

	# KULTUR - contributors
	for( qw( contributors corp_contributors ) )
	{
		next unless $eprint->exists_and_set( $_ );
		$table->appendChild( $session->render_row(
			$session->html_phrase( "eprint_fieldname_$_" ),
			$eprint->render_value( $_ ) ) );
	}

	#official_url
	if( $eprint->is_set( "official_url" ) )
	{
		$table->appendChild( $session->render_row(
					$session->html_phrase( "eprint_fieldname_official_url" ),
					$eprint->render_value( "official_url" ) ) );
	}

	# Commentary
	if( $eprint->is_set( "commentary" ) )
	{
		my $target = EPrints::DataObj::EPrint->new( 
			$session,
			$eprint->get_value( "commentary" ),
			$session->get_repository()->get_dataset( "archive" ) );
		if( defined $target )
		{
			$table->appendChild( $session->render_row(
				$session->html_phrase( 
					"eprint_fieldname_commentary" ),
				$target->render_citation_link() ) );
		}
	}

	my $frag = $session->make_doc_fragment;
	$frag->appendChild( $eprint->render_value( "type"  ) );
	my $type = $eprint->get_value( "type" );
	if( $type eq "conference_item" )
	{
		$frag->appendChild( $session->make_text( " (" ));
		$frag->appendChild( $eprint->render_value( "pres_type"  ) );
		$frag->appendChild( $session->make_text( ")" ));
	}
	if( $type eq "monograph" )
	{
		$frag->appendChild( $session->make_text( " (" ));
		$frag->appendChild( $eprint->render_value( "monograph_type"  ) );
		$frag->appendChild( $session->make_text( ")" ));
	}
	if( $type eq "thesis" )
	{
		$frag->appendChild( $session->make_text( " (" ));
		$frag->appendChild( $eprint->render_value( "thesis_type"  ) );
		$frag->appendChild( $session->make_text( ")" ));
	}
	$table->appendChild( $session->render_row(
		$session->html_phrase( "eprint_fieldname_type" ),
		$frag ));

	# Additional Info
	if( $eprint->is_set( "note" ) )
	{
		$table->appendChild( $session->render_row(
			$session->html_phrase( "eprint_fieldname_note" ),
			$eprint->render_value( "note" ) ) );
	}

	# Keywords
	if( $eprint->is_set( "keywords" ) )
	{
		$table->appendChild( $session->render_row(
			$session->html_phrase( "eprint_fieldname_keywords" ),
			$eprint->render_value( "keywords" ) ) );
	}

	# KULTUR - show additional metadata
	for( qw( divisions date copyright_holders funders id_number related_url_url related_exhibitions event_location projects physical_location location_dates material measurements ) )
	{
		next unless $eprint->exists_and_set( $_ );
		$table->appendChild( $session->render_row(
			$session->html_phrase( "eprint_fieldname_$_" ),
			$eprint->render_value( $_ ) ) );
	}

	$table->appendChild( $session->render_row(
		$session->html_phrase( "page:id_code" ),
		$eprint->render_value( "eprintid" ) ) );


	my $user = new EPrints::DataObj::User( 
			$eprint->{session},
 			$eprint->get_value( "userid" ) );
	my $usersname;
	if( defined $user )
	{
		$usersname = $user->render_description();
	}
	else
	{
		$usersname = $session->html_phrase( "page:invalid_user" );
	}

	$table->appendChild( $session->render_row(
		$session->html_phrase( "page:deposited_by" ),
		$usersname ) );

	if( $eprint->is_set( "datestamp" ) )
	{
		$table->appendChild( $session->render_row(
			$session->html_phrase( "page:deposited_on" ),
			$eprint->render_value( "datestamp" ) ) );
	}

	if( $eprint->is_set( "lastmod" ) )
	{
		$table->appendChild( $session->render_row(
			$session->html_phrase( "eprint_fieldname_lastmod" ),
			$eprint->render_value( "lastmod" ) ) );
	}

	return {
		id => "metadata",
		title => $session->html_phrase( "page:metadata" ),
		content => $table,
	};
}

# render the document request button
sub kultur_render_request
{
	my( $header, $footer, $eprint, $docs ) = @_;

	return if @$docs == 0 or $eprint->get_value( "eprint_status" ) ne "archive";

	my $session = $eprint->{session};

	# Contact email address
	my $has_contact_email = 0;
	if(
		$session->get_repository->can_call( "email_for_doc_request" ) &&
		defined( $session->get_repository->call( "email_for_doc_request", $session, $eprint ) )
	  )
	{
		$has_contact_email = 1;
	}

	return unless $has_contact_email;

	$header->appendChild( $session->html_phrase( "page:nofulltext" ) );

	# "Request a copy" button
	my $form = $session->render_form( "get", $session->get_repository->get_conf( "perl_url" ) . "/request_doc" );
	$header->appendChild( $form );
	$form->appendChild( $session->render_hidden_field( "eprintid", $eprint->get_id ) );
	$form->appendChild( $session->render_action_buttons( 
		"null" => $session->phrase( "request:button" )
	) );
}

# calculate rows/columns for a given number of documents and the thumbnail
# size to use
use POSIX;
sub kultur_rows_columns
{
	my( $session, $total ) = @_;

	my $kultur = $session->get_repository->get_conf( "kultur" );
	my $geoms = $kultur->{"geoms"};
	my $geom_keys = $kultur->{"geom_keys"};
	my $max_width = $kultur->{"max_width"};
	my $max_height = $kultur->{"max_height"};

	my( $rows, $cols, $size, $text );

	# calculate the number of rows/columns for each available thumbnail size
	# and find the best fit, where we can overflow the height but not the width
	for(@$geom_keys)
	{
		$size = $_;
		my $geom = $geoms->{$_};
		$text = int($geom->[0]/6); # max chars
		$cols = int($max_width / $geom->[0]);
		$rows = int(($total / $cols)+.5);
		last if $rows * $geom->[1] <= $max_height;
	}

	return( $rows, $cols, $size, $text );
}

# render the document tab(s)
sub kultur_render_documents
{
	my( $session, $dataset, $eprint, $docs ) = @_;

	# Contact email address
	my $has_contact_email = 0;
	if(
		$session->get_repository->can_call( "email_for_doc_request" ) &&
		defined( $session->get_repository->call( "email_for_doc_request", $session, $eprint ) )
	  )
	{
		$has_contact_email = 1;
	}

	my( $doctable, $doctr, $doctd );

	my $document_groups = {};

	# assign each document to the appropriate tab
	foreach my $doc (@$docs)
	{
		my $format = $doc->get_value('format');
		if ($doc->get_value('security') ne 'public')
		{
			push @{$document_groups->{secure}}, $doc;
		}
		elsif ($format =~ /^image/)
		{
			push @{$document_groups->{image}}, $doc;
		}
		elsif ($format =~ /^(video)|(audio)/)
		{
			push @{$document_groups->{media}}, $doc;
		}
		else
		{
			push @{$document_groups->{other}}, $doc;
		}

	}

	my @tabs;

	foreach my $document_type (qw/ image media secure other /)
	{
		next unless defined $document_groups->{$document_type};

		my $total = scalar(@{$document_groups->{$document_type}});
		next if $total == 0;

		$doctable = $session->make_element( "table", class=>"ep_block" );

		my( $rows, $columns, $size, $max_text ) =
			kultur_rows_columns( $session, $total );

		my $column_count = 0;
		foreach my $doc ( @{$document_groups->{$document_type}} )
		{
			if( $column_count == 0 || $column_count >= $columns )
			{
				$doctr = $session->make_element( "tr" );
				$doctable->appendChild( $doctr );
				$column_count = 0;
			}
			++$column_count;

			$doctd = $session->make_element( "td", valign=>"top", style=>"text-align:center" );
			$doctr->appendChild( $doctd );

			my $caption = $doc->render_citation_link( "lightbox" );
			my $lightbox_caption = $session->make_doc_fragment;
			$lightbox_caption->appendChild( $caption->cloneNode( 1 ) );
			$lightbox_caption->appendChild( $eprint->render_citation_link( "lightbox" ) );
			$lightbox_caption = EPrints::XML::to_string( $lightbox_caption );
			my $link_title = EPrints::Utils::tree_to_utf8( $caption );
			EPrints::XML::dispose( $caption );

			my $ddiv = $session->make_element( "span", class => "ep_lightbox_link", title => $lightbox_caption );
			$doctd->appendChild( $ddiv );

			my $link_caption;
			if( $doc->exists_and_set( "formatdesc" ) )
			{
				$link_caption = $doc->get_value( "formatdesc" );
				$link_caption =~ s/^(.{$max_text}).*/$1 .../;
				$link_caption = $session->make_text( $link_caption );
			}
			else
			{
				#$link_caption = $doc->render_value( "format" );
				$link_caption = $session->make_text( "Untitled" );
			}
			
			if ($document_type eq 'image' || $document_type eq 'media')
			{
				my $lightbox_url =
					kultur_thumbnail_url( $doc, "video_preview" )
					|| kultur_thumbnail_url( $doc, "audio_preview" )
					|| kultur_thumbnail_url( $doc, "lightbox" );


				if( !$doc->is_public )
				{
					$lightbox_url = undef;
				}

				$ddiv->appendChild( kultur_icon_link( $doc, size => $size, preview => 1, rel => "lightbox[icon_$document_type]", title => $link_title ) );

				$ddiv->appendChild( $session->make_element( 'br' ));
			
				if( $lightbox_url )
				{
					my $preview_link = $session->make_element('a', href=> $lightbox_url, rel => "lightbox[link_$document_type]", title => $link_title );
					$preview_link->appendChild( $link_caption );
					$ddiv->appendChild( $preview_link );
				}
				else
				{
					my $download_link = $session->make_element('a', href=> $doc->get_url, title => $link_title);
					$download_link->appendChild( $link_caption );
					$ddiv->appendChild( $download_link );
				}
			}
			else
			{
				$ddiv->appendChild( kultur_icon_link( $doc, size => $size, title => $link_title ) );

				$ddiv->appendChild( $session->make_element( 'br' ));
			
				if( $has_contact_email && !$doc->is_public && $eprint->get_value( "eprint_status" ) eq "archive" )
				{
# "Request a copy" button
					$doctd = $session->make_element( "td" );
					my $form = $session->render_form( "get", $session->get_repository->get_conf( "perl_url" ) . "/request_doc" );
					$form->appendChild( $session->render_hidden_field( "docid", $doc->get_id ) );
					$form->appendChild( $session->render_action_buttons( 
								"null" => $session->phrase( "request:button" )
								) );
					$doctd->appendChild( $form );
					$doctr->appendChild( $doctd );
				}
				else
				{
					my $download_link = $session->make_element('a', href=> $doc->get_url, title => $link_title);
					$download_link->appendChild( $link_caption );
					$ddiv->appendChild( $download_link );

					my %files = $doc->files;
					if( defined $files{$doc->get_main} )
					{
						my $size = $files{$doc->get_main};
						$ddiv->appendChild( $session->make_text( " (" . EPrints::Utils::human_filesize($size) . ")" ));
					}
				}
			}

		}

		my $title = $session->html_phrase( "document_group_name_" . $document_type );

		push @tabs, {
			id => $document_type,
			title => $title,
			content => $doctable,
		};
	}

	return @tabs;
}

# render the header <link>s
sub kultur_render_links
{
	my( $session, $dataset, $eprint ) = @_;

	my $links = $session->make_doc_fragment();

	$links->appendChild( $session->plugin( "Export::Simple" )->dataobj_to_html_header( $eprint ) );
	$links->appendChild( $session->plugin( "Export::DC" )->dataobj_to_html_header( $eprint ) );

	return $links;
}

sub kultur_icon_link
{
	my( $self, %opts ) = @_;

#adam added size default to support choice of size as an option
	$opts{size} = 'small' unless defined $opts{size};

	$opts{public} = 1 unless defined $opts{public};
	if( $opts{public} && !$self->is_public )
	{
		$opts{preview} = 0;
	}

	my %aopts;
	$aopts{href} = $self->get_url;
	$aopts{title} = $opts{title};
	$aopts{target} = "_blank" if( $opts{new_window} );

	my $preview_id = "doc_preview_".$self->get_id;

	# If we have a preview available link to it instead, using the lightbox
	if ($opts{preview} and $opts{rel} and $opts{rel} =~ m/^lightbox/)
	{
		my $lightbox_url =
			kultur_thumbnail_url( $self, "video_preview" )
			|| kultur_thumbnail_url( $self, "audio_preview" )
			|| kultur_thumbnail_url( $self, "lightbox" );
		if( $lightbox_url )
		{
			$aopts{href} = $lightbox_url;
			$aopts{rel} = $opts{rel};
		}
	} 

	if( !defined( $aopts{title} ) )
	{
		my $caption = $self->render_citation_link( "lightbox" );
		$aopts{title} = EPrints::Utils::tree_to_utf8( $caption );
		EPrints::XML::dispose( $caption );
	}

	my $a = $self->{session}->make_element( "a", %aopts );
	$a->appendChild( $self->{session}->make_element( 
		"img", 
		class=>"ep_doc_icon",
		alt=>"[img]",
#modified by adam to increase icon size
#		src=>$self->icon_url( public=>$opts{public} ),
		src=>$self->icon_url( public=>$opts{public},size=>$opts{size} ),
		border=>0 ));
	my $f = $self->{session}->make_doc_fragment;
	$f->appendChild( $a ) ;

	return $f;
}

sub kultur_thumbnail_url
{
	my( $doc, $resource ) = @_;

	$resource = "small" unless defined $resource;

	my $preview_doc = $doc->get_related_objects( EPrints::Utils::make_relation( "has".$resource."Version" ) )->[0];

	if (defined $preview_doc) {
		return $preview_doc->get_url;
	} else {
		return $doc->thumbnail_url($resource);
	}
}
