package EPrints::Plugin::Convert::Kultur;

=head1 NAME

EPrints::Plugin::Convert::Kultur - image and video conversion

=head1 SYNOPSIS

  use EPrints;

  ...

  my $plugin = $session->plugin( "Convert" );
  my %available = $plugin->can_convert( $doc );
  $plugin = $available{"thumbnail_video_preview"}->{plugin};
  $new_doc = $plugin->convert( $doc, "thumbnail_video_preview" );


=head1 DESCRIPTION

Conversion of images, videos and audio into preview/thumbnail versions.

This plugin wraps the ImageMagick I<convert> and I<ffmpeg> tools.

=head1 PARAMETERS

These parameters can be set through the B<plugins.pl> configuration file. You must also configure the B<executables> locations for I<convert> and I<ffmpeg> in L<EPrints::SystemSettings>.

=over 4

=item convert_formats = { ext => mime_type }

Define the formats supported for input by the call_convert() method.

=item ffmpeg_formats = { ext => mime_type }

Define the formats supported for input by the call_ffmpeg() method.

=item sizes = { size => [$w, $h] }

Define the size of thumbnails that can be generated e.g. "small => [66,50]". The image dimensions generated may be smaller than those specified if the aspect ratio of the source document is different.

Images are output in 8-bit paletted PNG.

=item video_preview = 1

Enable video previews (flash video).

=item audio_preview = 1

Enable audio previews (MP3 audio).

=item video_height = "480"

If generating a video preview use the number of vertical lines given.

=item audio_sampling = "22050"

Audio frequency sampling rate.

=item audio_bitrate = "64k"

Audio bit rate (in kilobits).

=item audio_codec = "libmp3lame"

Your I<ffmpeg> may be compiled with a different MP3 audio encoder to the standard "libmp3lame" (do C<ffmpeg -formats | grep mp3>).

=item frame_rate = "10.00"

Video frame rate.

=item video_rate = "1500k"

Video bit rate (in kilobits).

=item call_convert = sub( $plugin, $dst, $doc, $src, $geom, $size )

See L</call_convert>.

=item call_ffmpeg = sub( $plugin, $dst, $doc, $src, $geom, $size, $offset )

See L</call_ffmpeg>.

=back

=head1 METHODS

=over 4

=cut

use EPrints::Plugin::Convert;
@ISA = qw/ EPrints::Plugin::Convert /;

use strict;

# default settings
my %DEFAULT;

# formats supported by ImageMagick
$DEFAULT{convert_formats} = {qw(
	bmp image/bmp
	gif image/gif
	ief image/ief
	jpeg image/jpeg
	jpe image/jpeg
	jpg image/jpeg
	png image/png
	tiff image/tiff
	tif image/tiff
	pnm image/x-portable-anymap
	pbm image/x-portable-bitmap
	pgm image/x-portable-graymap
	ppm image/x-portable-pixmap
)};
# formats supported by ffmpeg
$DEFAULT{ffmpeg_formats} = {qw(
	mp3	audio/mpeg
	mp4	video/mpeg
	mp2	video/mpeg
	mpa	video/mpeg
	mpe	video/mpeg
	mpeg	video/mpeg
	mpg	video/mpeg
	mpv2	video/mpeg
	mov	video/quicktime
	qt	video/quicktime
	lsf	video/x-la-asf
	lsx	video/x-la-asf
	asf	video/x-ms-asf
	asr	video/x-ms-asf
	avi	video/x-msvideo
	vob	video/mpeg
	m4v	video/mpeg
	m2v	video/mpeg
	wmv	video/x-ms-wmv
	ogv	video/ogg
	flv	video/x-flv
	wav	audio/wav
	ac3	audio/ac3
	m4a	audio/mp4
)};
# thumbnail sizes to output
$DEFAULT{sizes} = {(
	small => [66,50],
	medium => [200,150],
	preview => [400,300],
#	lightbox => [640,480],
)};
# enable/disable audio/video previews
$DEFAULT{video_preview} = 1;
$DEFAULT{audio_preview} = 1;

# video_preview lines
$DEFAULT{video_height} = "480";

# audio_preview frequency
$DEFAULT{audio_sampling} = "22050";

# ffmpeg quality settings
$DEFAULT{audio_codec} = "libmp3lame";
$DEFAULT{audio_bitrate} = "64k";
$DEFAULT{frame_rate} = "10.00";
$DEFAULT{video_rate} = "1500k";

# methods
$DEFAULT{call_convert} = \&call_convert;
$DEFAULT{call_ffmpeg} = \&call_ffmpeg;

sub new
{
	my( $class, %opts ) = @_;

	my $self = $class->SUPER::new( %opts );

	$self->{name} = "Image thumbnails";
	$self->{visible} = "all";
	for(qw(
		video_preview audio_preview 
		convert_formats ffmpeg_formats sizes
		audio_sampling audio_bitrate frame_rate video_rate video_height audio_codec time_offset
		call_convert call_ffmpeg
	  ))
	{
		# defined by new( foo => bar )
		next if defined $self->{$_};
		# defined by $c->{plugins}
		if( defined $self->{session} )
		{
			$self->{$_} = $self->param( $_ );
		}
		# set the default
		$self->{$_} = $DEFAULT{$_} if !defined $self->{$_};
	}
	if( defined $self->{session} )
	{
		my $cfg = $self->{session}->get_repository->get_conf( "executables" );
		if( !defined( $self->{'convert'} = $cfg->{'convert'} ) )
		{
			$self->{'convert_formats'} = {};
		}
		if( !defined( $self->{'ffmpeg'} = $cfg->{'ffmpeg'} ) )
		{
			$self->{'ffmpeg_formats'} = {};
		}
	}

	return $self;
}

=item $ver = $plugin->convert_version()

Returns the MAJOR.MINOR version of ImageMagick.

Returns 0.0 if the version can not be determined.

=cut

sub convert_version
{
	my( $self ) = @_;

	return 0.0 if !defined $self->{convert};

	my $version = `$self->{convert} --version`;
	$version =~ s/^.*Version: ImageMagick ([0-9]+\.[0-9]+)\..*$/$1/s;

	return $version || 0;
}

=item $ok = $plugin->is_video( $doc )

Returns true if $doc is a video.

=cut

sub is_video
{
	my( $self, $doc ) = @_;

	return $doc->exists_and_set( "media_video_codec" );
}

=item %types = $plugin->can_convert( $doc )

Returns a hash map of types this plugin can convert $doc to.

This may be relatively expensive to do if the plugin has to call an external tool to determine if it can export something.

=cut

sub can_convert
{
	my( $self, $doc ) = @_;

	return () unless $self->get_repository->can_execute( "convert" );

	my %types;

	# Get the main file name
	my $fn = $doc->get_main() or return ();

	my( $ext ) = $fn =~ /\.([^\.]+)$/;
	return () unless defined $ext;

	if( exists $self->{convert_formats}->{lc($ext)} )
	{
		foreach my $size (keys %{$self->{sizes}})
		{
			$types{"thumbnail_$size"} = { plugin => $self };
		}
	}
	if( exists $self->{ffmpeg_formats}->{lc($ext)} )
	{
		if( $self->{video_preview} && $doc->exists_and_set( "media_video_codec" ) )
		{
			foreach my $size ("video_preview", keys %{$self->{sizes}})
			{
				$types{"thumbnail_$size"} = { plugin => $self };
			}
		}
		# only offer audio thumbnailing if there's no video
		elsif( $self->{audio_preview} && $doc->exists_and_set( "media_audio_codec" ) )
		{
			$types{"thumbnail_audio_preview"} = { plugin => $self };
		}
	}

	return %types;
}

=item $new_doc = $plugin->convert( $eprint, $doc, $type )

Request the plugin converts $doc to $type, as returned by L</can_convert>.

=cut

sub convert
{
	my( $self, $eprint, $doc, $type ) = @_;

	my $new_doc = $self->SUPER::convert( $eprint, $doc, $type );

	return undef if !defined $new_doc;

	$new_doc->set_value( "format", $self->{_mime_type} );

	return $new_doc;
}

=item @filelist = $plugin->export( $dir, $doc, $type )

Request the plugin converts $doc to $type, as returned by L</can_convert>. Outputs the resulting files to $dir and returns their paths (excluding the leading $dir part).

=cut

sub export
{
	my ( $self, $dir, $doc, $type ) = @_;
	
	my $doc_src = $doc->get_stored_file( $doc->get_main );
        my $src = $doc_src->get_local_copy;

	return () unless -r $src and -s $src;

	my( $size ) = $type =~ m/^thumbnail_(.*)$/;
	return () unless defined $size;
	my $geom;
	if( $size ne "audio_preview" && $size ne "video_preview" )
	{
		$geom = $self->{sizes}->{$size};
		return () unless defined $geom;
	}
	
	my $fn;
	if( $size eq "audio_preview" )
	{
		$self->{_mime_type} = "video/x-flv";
		$fn = "$size.flv";
	}
	elsif( $size eq "video_preview" )
	{
		$self->{_mime_type} = "video/x-flv";
		$fn = "$size.flv";
	}
	else
	{
		$self->{_mime_type} = "image/png";
		$fn = "$size.png";
	}

	my $dst = "$dir/$fn";

	my( $ext ) = $src =~ /\.([^\.]+)$/;
	if( exists $self->{ffmpeg_formats}->{lc($ext)} )
	{
		my $seconds = 5; # default to 5 seconds (as good a place as any)
		my $duration = 0;
		if( $doc->exists_and_set( "media_sample_start" ) )
		{
			$seconds = $doc->get_value( "media_sample_start" );
			$duration = $doc->get_value( "media_duration" );
		}
		$seconds = $self->calculate_offset( $duration, $seconds );

		&{$self->{call_ffmpeg}}( $self, $dst, $doc, $src, $geom, $size, $seconds );
	}
	else
	{
		&{$self->{call_convert}}( $self, $dst, $doc, $src, $geom, $size );
	}

	unless( -s $dst ) {
		unlink( $dst );
		return ();
	}

	EPrints::Utils::chown_for_eprints( $dst );
	
	return ($fn);
}

=back

=head2 Utility Methods

=over 4

=item $plugin->call_convert( $dst, $doc, $src, $geom, $size )

Calls the ImageMagick I<convert> tool to convert $doc into a thumbnail image. Writes the image to $dst. $src is the full path to the main file from $doc. The resulting image should not exceed $geom dimensions ([w,h] array ref).

$size is the thumbnail-defined size (as-in the keys to the B<sizes> parameter).

This method can be overridden with the B<call_convert> parameter.

=cut

sub call_convert
{
	my( $self, $dst, $doc, $src, $geom, $size ) = @_;

	my $convert = $self->{'convert'};
	my $version = $self->convert_version;

	if (!defined($geom)) {
		EPrints::abort("NO GEOM");
	}

	$geom = "$geom->[0]x$geom->[1]";
# PNG8
	if( $size eq "small" && $version > 6.3 ) # geom^ requires 6.3.8
	{
		$self->_system($convert, "-strip", "-colorspace", "RGB", "-background", "white", "-thumbnail","$geom^", "-gravity", "center", "-extent", $geom, $src."[0]", "PNG8:$dst");
	}
	elsif( $size eq "medium" )
	{
		$self->_system($convert, "-strip", "-colorspace", "RGB", "-bordercolor", "white", "-border", "1x1", "-trim", "+repage", "-size", "$geom", "-thumbnail","$geom>", "-background", "white", "-gravity", "center", "-extent", $geom, $src."[0]", "PNG8:$dst");
	}
	else
	{
		$self->_system($convert, "-strip", "-colorspace", "RGB", "-background", "white", "-thumbnail","$geom>", "-extract", $geom, $src."[0]", "PNG8:$dst");
	}
#	$self->_system($convert, "-strip", "-colorspace", "RGB", "-thumbnail","$geom>", "-extract", $geom, $src."[0]", "$dst");

	return -s $dst;
}

=item $plugin->call_ffmpeg( $dst, $doc, $src, $geom, $size, $offset )

Uses the I<ffmpeg> tool to do the image conversion of $doc. $dst is the filename to write to, $src is the filename to read from and $geom is the image dimensions to write.

$size is the thumbnail-defined size (as-in the keys to the B<sizes> parameter or B<audio_preview> or B<video_preview>).

$offset is the time offset to extract (for B<audio_preview>/B<video_preview>). It is an array ref of [HOUR, MINUTE, SECOND, FRAME].

=cut

sub call_ffmpeg
{
	my( $self, $dst, $doc, $src, $geom, $size, $offset ) = @_;

	my $rc = 0;

	if( $size eq "audio_preview" )
	{
		$rc = $self->export_audio_flv( $dst, $doc, $src, $self->{audio_sampling} );
	}
	elsif( $size eq "video_preview" )
	{
		$rc = $self->export_flv( $dst, $doc, $src, $self->{video_height} );
	}
	else
	{
		$rc = $self->export_cell( $dst, $doc, $src, $geom, $size, $offset );
	}

	return $rc;
}

=item $plugin->export_mp3( $dst, $doc, $src, $rate )

Export $src to $dst in MP3 format at sampling rate $rate.

=cut

sub export_mp3
{
	my( $self, $dst, $doc, $src, $rate ) = @_;

	my $ffmpeg = $self->{ffmpeg};

	my $cmd = sprintf("%s -y -i %s -acodec libmp3lame -ac 2 -ar %s -ab 32 -f mp3 %s", $ffmpeg, quotemeta($src), $rate, quotemeta($dst));
	$self->_system("$cmd >/dev/null 2>&1");

	unless( -s $dst )
	{
		print STDERR Carp::longmess( "Error in command: $cmd" );
		return;
	}
}

=item $plugin->export_audio_flv( $dst, $doc, $src, $rate )

Export audio-only $src to $dst in Flash Video format.

=cut

sub export_audio_flv
{
	my( $self, $dst, $doc, $src, $rate ) = @_;

	my $ffmpeg = $self->{ffmpeg};

	return unless $doc->exists_and_set( "media_audio_codec" );

	my $cmd = sprintf('%s -y -i %s -acodec libmp3lame -ac 2 -ar %s -ab %s -f flv %s ', $ffmpeg, quotemeta($src), $rate, $self->{audio_bitrate}, quotemeta($dst));
#print STDERR "$cmd\n";
	$self->_system("$cmd >/dev/null 2>&1");

	unless( -s $dst )
	{
		print STDERR Carp::longmess( "Error in command: $cmd" );
		return;
	}

	return 1;
}

=item $plugin->export_flv( $dst, $doc, $src, $lines )

Export $src to $dst in Flash Video format with vertical lines $lines.

=cut

sub export_flv
{
	my( $self, $dst, $doc, $src, $lines ) = @_;

	my $ffmpeg = $self->{ffmpeg};

	return unless $doc->exists_and_set( "media_video_codec" );

	my $width = $doc->get_value( "media_width" );
	my $height = $doc->get_value( "media_height" );
	my $duration = $doc->get_value( "media_duration" );
	my $aspect = $doc->get_value( "media_aspect_ratio" );
	my $ratio = 4 / 3;
	if( defined $aspect && $aspect =~ /^(\d+):(\d+)$/ )
	{
		$ratio = $1 / $2;
	}
	elsif( defined $width && defined $height )
	{
		$ratio = $width / $height;
	}
#print STDERR "Aspect ratio = $ratio\n";

	if( defined $height && $height > $lines )
	{
		$height = $lines;
	}
	elsif( !defined $height )
	{
		$height = $lines;
	}
# width must be a multiple of 2 or ffmpeg barfs
	$width = sprintf("%.0f", $height * $ratio );
	$width++ if $width % 2;
#print STDERR "Outputting $width x $height\n";

	my $cmd = sprintf('%s -y -i %s -acodec libmp3lame -ac 2 -ar %s -ab %s -f flv -r %s -s %sx%s -b %s %s ', $ffmpeg, quotemeta($src), $self->{audio_sampling}, $self->{audio_bitrate}, $self->{frame_rate}, $width, $height, $self->{video_rate}, quotemeta($dst));
#print STDERR "$cmd\n";
	$self->_system("$cmd >/dev/null 2>&1");

	unless( -s $dst )
	{
		print STDERR Carp::longmess( "Error in command: $cmd" );
		return;
	}

	return 1;
}

=item $plugin->export_cell( $dst, $doc, $src, $geom, $size, $offset )

Export $src to $dst in PNG8 format in dimensions $geom from offset $offset.

=cut

sub export_cell
{
	my( $self, $dst, $doc, $src, $geom, $size, $offset ) = @_;
	
	my $ffmpeg = $self->{ffmpeg};

	my $tmpfile;
	if( $dst =~ /\.jpg$/i )
	{
		$tmpfile = $dst;
	}
	else
	{
		$tmpfile = File::Temp->new( SUFFIX => ".jpg" );
	}

	$offset = _seconds_to_marker( $offset );

	# extract the frame in MJPEG format
	#my $cmd = sprintf('%s -y -i %s -vframes 1 -an -f rawvideo -ss %s -vcodec mjpeg -s %sx%s %s', $ffmpeg, quotemeta($src), $offset, @$geom, quotemeta($tmpfile));
	my $cmd = sprintf('%s -i %s -an -f mjpeg -ss %s -t 00:00:01 -r 1 -y -s %sx%s %s', $ffmpeg, quotemeta($src), $offset, @$geom, quotemeta($tmpfile));
	$self->_system("$cmd >/dev/null 2>&1");

	unless( -s $tmpfile )
	{
		print STDERR Carp::longmess( "Error in command: $cmd" );
		return;
	}

	# don't need to further convert
	return 1 if $tmpfile eq $dst;

	# now use call_convert to generate the final thumbnail
	return &{$self->{call_convert}}( $self, $dst, $doc, "$tmpfile", $geom, $size );
}

=item $settings = $plugin->get_dimensions( $src )

Returns vital information about a media file $src.

=cut

sub get_dimensions
{
	my( $self, $src ) = @_;

	my %settings = ( streams => {} );

	my $ffmpeg = $self->{ffmpeg};

	my $ffmpeg_log = File::Temp->new;
	my $cmd = "$ffmpeg -i ".quotemeta($src);
	$self->_system("$cmd 2>$ffmpeg_log");
	seek($ffmpeg_log,0,0);

	while(<$ffmpeg_log>)
	{
		if( /Stream #[0-9.]+[^:]*?: (Audio|Video): / )
		{
			$settings{streams}->{"\L$1"} = 1;
		}
		if( /Video: (\S+),.*?\s(\d+)x(\d+)\b/ )
		{
			$settings{video_codec} = $1;
			$settings{width} = $2;
			$settings{height} = $3;
			$settings{video_codec} =~ s/(['"])(.+)\1/$2/;
		}
		if( /Audio: (\S+),/ )
		{
			$settings{audio_codec} = $1;
			$settings{audio_codec} =~ s/(['"])(.+)\1/$2/;
		}
		if( /Duration:.* (\d+):(\d\d):(\d\d)\.\d+/ )
		{
			$settings{duration} = $1 * 3600 + $2 * 60 + $3;
		}
		if( /Video:.* DAR (\d+:\d+)\b/ )
		{
			$settings{aspect_ratio} = $1;
		}
	}

	return \%settings;
}

=item $secs = $plugin->calculate_offset( $duration, $offset )

Translates a seconds or percentage offset into seconds from the start time. If the resulting time is greater than $duration returns $duration.

To specify seconds either use just a number (1234) or append 's' (1234s).

To specify a percentage of $duration append '%' (52%).

=cut

sub calculate_offset
{
	my( $self, $duration, $offset ) = @_;

	$duration = 0 if !defined $duration;
	$offset = "" if !defined $offset;

	my $seconds = 0;
	if( $offset =~ /([0-9]+)s/ )
	{
		$seconds = $1;
		$seconds = $duration if $seconds > $duration;
	}
	elsif( $offset =~ /([0-9\.]+)\%/ )
	{
		$seconds = $duration * ($1/100);
		$seconds = $duration if $seconds > $duration;
	}
	elsif( $offset >= 0 && int($offset) <= $duration ) # ignore .frames
	{
		$seconds = $offset+0;
	}
	elsif( $duration >= 10 )
	{
		$seconds = 10;
	}
	elsif( $duration >= 5 )
	{
		$seconds = 5;
	}
	return $seconds;
}

sub _seconds_to_marker
{
	my( $seconds ) = @_;

	# enforce numeric
	$seconds = 0+$seconds;
	my $frames = $seconds-int($seconds);
	$seconds = int($seconds);

	my @offset = (int($seconds/3600), int(($seconds%3600)/60), int($seconds) % 60, $frames);

	return join(':', @offset[0..2]).".".$offset[3];
}

# make a system call
sub _system
{
	my( $self, @args ) = @_;

	my $rc = system(@args);

	return $rc;
}

=back

=head1 SEE ALSO

L<EPrints::Plugin>, L<EPrints::Plugin::Convert>.

=head1 AUTHOR

Copyright 2009 Tim Brody <tdb2@ecs.soton.ac.uk>, University of Southampton, UK.

This module is released under the GPLv3 license.

=cut

1;
