#!/usr/bin/perl

=head1 NAME

build_ffmpeg.pl - download and build ffmpeg for Kultur

=head1 SYNOPSIS

B<build_ffmpeg.pl> [B<options>]

=head1 DESCRIPTION

This tool builds ffmpeg from source to provide additional format support.

It downloads ffmpeg and dependent library sources.

Compiles dependent libraries into a /usr structure below the current path.

Builds ffmpeg and ffprobe.

=head1 OPTIONS

=over 8

=back

=cut

use Getopt::Long;
use Pod::Usage;
use LWP::UserAgent;
use DB_File;

use strict;

my $opt_verbose = 0;
my $opt_help = 0;

GetOptions(
	"help" => \$opt_help,
	"verbose+" => \$opt_verbose,
) or pod2usage( 2 );

my $OK_OFFSET = 70;

tie my %MAKEFILE, 'DB_File', "MAKEFILE.DB";

my $root = `pwd`;
chomp($root);
my $prefix = "$root/usr";
my $libprefix = "$prefix/lib";
my $includeprefix = "$prefix/include";
mkdir($prefix);
for(qw( bin lib sbin include share))
{
	mkdir("$prefix/$_");
}

my $LOGFILE = "$root/build_ffmpeg.log";
unlink($LOGFILE);

my %bins = (
	svn => { name => "Subversion Version Control Client", required => 1, },
	git => { name => "Git Version Control Client", },
	gcc => { name => "GCC compiler", required => 1, },
	make => { name => "Automake make", required => 1, },
	md5sum => { name => "MD5 checksum tool", required => 1, },
	yasm => { name => "YASM assembly compiler", required => 1, },
);

print "Checking tool dependencies:\n";

my $ok = 1;
foreach my $bin (sort keys %bins)
{
	my $name = $bins{$bin}->{name};
	print sprintf("%s%s", $name, "."x(70-length($name)));
	my $path = which($bin);
	if( !$path )
	{
		$ok = 0 if $bins{$bin}->{required};
		not_ok();
	}
	else
	{
		ok();
	}
}

die "Failed: missing required dependencies\n" if !$ok;

my $rc;
my $ua = LWP::UserAgent->new();

my %libs = (
	"faac-1.28" => {
		name => "Freeware Advanced Audio Coder (faac)",
		filename => "faac-1.28.tar.gz",
		url => "http://downloads.sourceforge.net/project/faac/faac-src/faac-1.28/faac-1.28.tar.gz",
		md5sum => "80763728d392c7d789cde25614c878f6",
	},
	"faad2-2.7" => {
		name => "Freeware Advanced Audio Coder (faad2)",
		filename => "faad2-2.7.tar.gz",
		url => "http://downloads.sourceforge.net/project/faac/faad2-src/faad2-2.7/faad2-2.7.tar.gz",
		md5sum => "ee1b4d67ea2d76ee52c5621bc6dbf61e",
	},
	"libogg-1.1.4" => {
		name => "Ogg Audio Coder",
		filename => "libogg-1.1.4.tar.gz",
		url => "http://downloads.xiph.org/releases/ogg/libogg-1.1.4.tar.gz",
		md5sum => "10200ec22543841d9d1c23e0aed4e5e9",
	},
	"libvorbis-1.2.3" => {
		name => "Vorbis Media Container",
		filename => "libvorbis-1.2.3.tar.gz",
		url => "http://downloads.xiph.org/releases/vorbis/libvorbis-1.2.3.tar.gz",
		md5sum => "5aa77f55c0e0aab8eb8ed982335daac8",
	},
	"speex-1.2rc1" => {
		name => "SPEEX Audio Coder",
		filename => "speex-1.2rc1.tar.gz",
		url => "http://downloads.xiph.org/releases/speex/speex-1.2rc1.tar.gz",
		md5sum => "c4438b22c08e5811ff10e2b06ee9b9ae",
	},
	"libtheora-1.0" => {
		name => "Theora Video Coder",
		filename => "libtheora-1.0.tar.bz2",
		url => "http://downloads.xiph.org/releases/theora/libtheora-1.0.tar.bz2",
		md5sum => "c963937053f45a7878954bed37ceb182",
	},
	"schroedinger-1.0.7" => {
		name => "Schroedinger Video Codec (DIRAC)",
		filename => "schroedinger-1.0.7.tar.gz",
		url => "http://diracvideo.org/download/schroedinger/schroedinger-1.0.7.tar.gz",
		md5sum => "9a1132c54a58a6a040514f754b53ee44",
		depends => [qw( liboil-0.3.16 )],
	},
	"liboil-0.3.16" => {
		name => "liboil",
		filename => "liboil-0.3.16.tar.gz",
		url => "http://liboil.freedesktop.org/download/liboil-0.3.16.tar.gz",
		md5sum => "febb1d9f9bc4c440fcf622dc90f8b6b7",
	},
	"x264-snapshot-20090824-2245" => {
		name => "x264 Audio Coder",
		filename => "x264-snapshot-20090824-2245.tar.bz2",
		url => "ftp://ftp.videolan.org/pub/videolan/x264/snapshots/x264-snapshot-20090824-2245.tar.bz2",
		md5sum => "fa4118e6f81301c40d7ad3668fa3a08d",
	},
	"xvidcore/build/generic" => {
		name => "XVIDCore Library",
		filename => "xvidcore-1.2.2.tar.bz2",
		url => "http://downloads.xvid.org/downloads/xvidcore-1.2.2.tar.bz2",
		md5sum => "6a3473a12c8a1fa7bdc2b5cb829fab58",
	},
# 3.98-2 causes an audio encoding error, so try to avoid mp3!
	"lame-398-2" => {
		name => "LAME MP3 Audio Coder",
		filename => "lame-398-2.tar.gz",
		url => "http://downloads.sourceforge.net/project/lame/lame/3.98.2/lame-398-2.tar.gz",
		md5sum => "719dae0ee675d0c16e0e89952930ed35",
	},
	"autoconf-2.64" => {
		name => "Autoconf 2.60+",
		filename => "autoconf-2.64.tar.gz",
		url => "http://ftp.gnu.org/gnu/autoconf/autoconf-2.64.tar.gz",
		md5sum => "30a198cef839471dd4926e92ab485361",
	},
	"libpng-1.2.39" => {
		name => "libpng 1.2.39",
		filename => "libpng-1.2.39.tar.gz",
		url => "http://downloads.sourceforge.net/project/libpng/00-libpng-stable/1.2.39/libpng-1.2.39.tar.gz",
		md5sum => "ddfeaf19b690985910c42e41974e8d65",
	},
	"jpeg-7" => {
		name => "libjpeg v7",
		filename => "jpegsrc.v7.tar.gz",
		url => "http://www.ijg.org/files/jpegsrc.v7.tar.gz",
		md5sum => "382ef33b339c299b56baf1296cda9785",
	},
	"tiff-3.9.1" => {
		name => "libtiff 3.9.1",
		filename => "tiff-3.9.1.tar.gz",
		url => "ftp://ftp.remotesensing.org/pub/libtiff/tiff-3.9.1.tar.gz",
		md5sum => "63c59a44f34ae0787f2d71de3d256e20",
	},
	"ImageMagick-6.5.5-7" => {
		name => "ImageMagick 6.5.5",
		filename => "ImageMagick-6.5.5-7.tar.gz",
		url => "ftp://ftp.imagemagick.org/pub/ImageMagick/ImageMagick-6.5.5-7.tar.gz",
		md5sum => "ba1c7c0369d634e024078f565a85507c",
		depends => [qw( libpng-1.2.39 jpeg-7 tiff-3.9.1 )],
		configure => join(" ", qw(
./configure --prefix=$PREFIX --without-perl
)),
	},
);

my @ffmpeg_dependencies = sort keys %libs;

$libs{ffmpeg} = {
	name => "ffmpeg",
	url => "svn://svn.ffmpeg.org/ffmpeg/trunk",
	depends => \@ffmpeg_dependencies,
	configure => join(" ", qw(
./configure
	--prefix=$PREFIX
	--enable-nonfree
	--enable-memalign-hack
	--enable-libxvid
	--enable-libx264
	--enable-libfaac
	--enable-libfaad
	--enable-libmp3lame
	--enable-libvorbis
	--enable-libtheora
	--enable-libspeex
	--enable-libschroedinger
	--enable-pthreads
	--enable-gpl
	--enable-postproc
	--enable-static
	--disable-shared
	--extra-cflags=--static
	--disable-ffplay
	--disable-ffserver
	--extra-ldflags=-L$LIBPREFIX
	--extra-cflags=-I$INCLUDEPREFIX
	--extra-libs=$LIBPREFIX/libfaac.a
	--extra-libs=$LIBPREFIX/libfaad.a
	--extra-libs=$LIBPREFIX/libmp3lame.a
	--extra-libs=$LIBPREFIX/libmp4ff.a
	--extra-libs=$LIBPREFIX/libogg.a
	--extra-libs=$LIBPREFIX/libx264.a
	--extra-libs=$LIBPREFIX/libxvidcore.a
	--extra-libs=$LIBPREFIX/libogg.a
	--extra-libs=$LIBPREFIX/libtheora.a
	--extra-libs=$LIBPREFIX/libvorbis.a
	--extra-libs=$LIBPREFIX/libspeex.a
	--extra-libs=$LIBPREFIX/liboil-0.3.a
	--extra-libs=$LIBPREFIX/liboil-0.3.so.0
	--extra-libs=$LIBPREFIX/libschroedinger-1.0.a
	)),
};

# build shared ffmpeg libraries
$libs{ffmpeg_shared} = {
	name => "ffmpeg",
	url => "svn://svn.ffmpeg.org/ffmpeg/trunk",
	depends => \@ffmpeg_dependencies,
	configure => join(" ", qw(
./configure
	--prefix=$PREFIX
	--enable-nonfree
	--enable-memalign-hack
	--enable-libxvid
	--enable-libx264
	--enable-libfaac
	--enable-libfaad
	--enable-libmp3lame
	--enable-libvorbis
	--enable-libtheora
	--enable-libspeex
	--enable-libschroedinger
	--enable-pthreads
	--enable-gpl
	--enable-postproc
	--enable-shared
	--disable-static
	--disable-ffplay
	--disable-ffserver
	--extra-ldflags=-L$LIBPREFIX
	--extra-cflags=-I$INCLUDEPREFIX
	--extra-libs=$LIBPREFIX/libfaac.a
	--extra-libs=$LIBPREFIX/libfaad.a
	--extra-libs=$LIBPREFIX/libmp3lame.a
	--extra-libs=$LIBPREFIX/libmp4ff.a
	--extra-libs=$LIBPREFIX/libogg.a
	--extra-libs=$LIBPREFIX/libx264.a
	--extra-libs=$LIBPREFIX/libxvidcore.a
	--extra-libs=$LIBPREFIX/libogg.a
	--extra-libs=$LIBPREFIX/libtheora.a
	--extra-libs=$LIBPREFIX/libvorbis.a
	--extra-libs=$LIBPREFIX/libspeex.a
	--extra-libs=$LIBPREFIX/liboil-0.3.a
	--extra-libs=$LIBPREFIX/liboil-0.3.so.0
	--extra-libs=$LIBPREFIX/libschroedinger-1.0.a
	)),
};

# can't build static version of ffprobe
$libs{"ffprobe"} = {
	name => "ffprobe",
	url => "https://ffprobe.svn.sourceforge.net/svnroot/ffprobe/trunk",
	depends => [qw( ffmpeg_shared ffmpeg )],
	configure => join(" ", qw(
./configure
	--prefix=$PREFIX
	)),
};

foreach my $lib (sort keys %libs)
{
	title($libs{$lib}->{name});
	my $url = $libs{$lib}->{url};
	my $filename = $libs{$lib}->{filename};
	if( !defined $filename )
	{
		if( -d $lib )
		{
			system_catch( $lib, ["svn", "update", $lib] );
		}
		else
		{
			system_catch( $lib, ["svn", "checkout", $url, $lib] );
		}
	}
	else
	{
		if( -e $filename )
		{
			print "Using existing '$filename'\n";
		}
		else
		{
			print "Fetching $url\n";
			my $r = $ua->get( $url,
				":content_file" => $filename,
			);
			if( $r->is_error )
			{
				die "Error: ".$r->as_string."\n";
			}
		}
		my $md5sum = $libs{$lib}->{md5sum};
		if( $md5sum )
		{
			my $cur = `md5sum $filename`;
			$cur =~ s/[\* ].*$//s;
			die "MD5 checksum '$cur' does not match reference '$md5sum' - you may need to remove '$filename' to re-attempt download. If this error persists then it is possible someone is attempting to inject a compromised version of $lib.\n" if $cur ne $md5sum;
		}
		
		if( !-d $lib )
		{
			print "Unpacking $filename\n";
			if( $filename =~ /\.gz$/ or $filename =~ /\.tgz$/ )
			{
				system_catch("Extract $lib",["tar","xzf",$filename]);
			}
			else
			{
				system_catch("Extract $lib",["tar","xjf",$filename]);
			}
		}
	}
}

# configure and compile libraries
my @lib_order = build_dependencies( "ffprobe", \%libs, {} );

$ENV{LD_RUN_PATH} = "$prefix/lib";
$ENV{LDFLAGS} = "-L$prefix/lib";
$ENV{CFLAGS} = "-I$prefix/include";
$ENV{CPPFLAGS} = "-I$prefix/include";
$ENV{PKG_CONFIG_PATH} = "$prefix/lib/pkgconfig";
$ENV{PATH} = "$prefix/bin:$ENV{PATH}"; # autoconf

foreach my $lib (@lib_order)
{
	chdir("$root/$lib");
	my %OLD_ENV;
	while(my( $key, $value ) = each %{$libs{$lib}->{env}||{}})
	{
		$OLD_ENV{$key} = $ENV{$key};
		$ENV{$key} = $value;
	}
	# attempt to make configure if it doesn't already exist
	if( !-s "configure" )
	{
		system_catch( $lib, ["autoreconf"] );
	}
	my $cur = md5sum( "Makefile" );
	my $configured = 0;
	if( !defined $MAKEFILE{$lib} or !-s "Makefile" or $cur ne $MAKEFILE{$lib} )
	{
		RECONFIGURE:
		my $configure =
			$libs{$lib}->{configure}
			|| "./configure --prefix=\$PREFIX --enable-static --enable-shared";
		$configure =~ s/\$PREFIX/$prefix/g;
		$configure =~ s/\$LIBPREFIX/$libprefix/g;
		$configure =~ s/\$INCLUDEPREFIX/$includeprefix/g;
		system_catch( "$lib", [$configure] );
		$MAKEFILE{$lib} = md5sum( "Makefile" );
		$configured = 1;
	}
	eval { system_catch( "$lib", ["make install"] ) };
	goto RECONFIGURE if( $@ && !$configured );
	die $@ if $@;
	while(my( $key, $value ) = each %OLD_ENV)
	{
		$ENV{$key} = $value;
	}
}

chdir($root);

sub build_dependencies
{
	my( $lib, $libs, $seen ) = @_;

	return () if $seen->{$lib};
	$seen->{$lib} = 1;

	my @out;

	foreach my $dep (@{$libs->{$lib}->{depends}||[]})
	{
		push @out, build_dependencies( $dep, $libs, $seen );
	}

	push @out, $lib;

	return @out;
}

sub title
{
	print "\n$_[0]\n".('-' x length($_[0]))."\n";
}

sub ok
{
	print "[OK]\n";
}

sub not_ok
{
	print "[FAILED]\n";
}

sub which
{
	my( $bin ) = @_;
	for(split /:/, $ENV{PATH})
	{
		if( -e "$_/$bin" )
		{
			return 1;
		}
	}
	return 0;
}

sub system_catch
{
	my( $action, $cmd ) = @_;

	if( @$cmd > 1 )
	{
		print "$action: ".join(" ",map { quotemeta($_) } @$cmd)."\n";
	}
	else
	{
		print "$action: $cmd->[0]\n";
	}
	open(OLD_STDOUT, ">&STDOUT") or die "Failed to save STDOUT";
	open(STDOUT, ">>$LOGFILE") or die "Failed to open log file $LOGFILE: $!";
	open(OLD_STDERR, ">&STDERR") or die "Failed to save STDERR";
	open(STDERR, ">>$LOGFILE") or die "Failed to open log file $LOGFILE: $!";
	my $rc = system(@$cmd);
	open(STDERR, ">&OLD_STDERR") or die "Failed to restore STDERR";
	open(STDOUT, ">&OLD_STDOUT") or die "Failed to restore STDOUT";
	die "Failed: $action\n" if $rc != 0;
}

sub md5sum
{
	my( $file ) = @_;

	my $md5 = `md5sum $file`;
	$md5 =~ s/[\* ].*$//sg;

	return $md5 || "";
}
